/* To compile use:
cc lex.c stringI.o -lm
*/

#include "scanner.h" /* Useful Macro Definitions */
#include <values.h>                 /* max. int. and max/min real num.
				       defined in values.h */
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

#define PERROR_1                 0
#define NOT_FOUND_1              -2  /* used in binary search */

#define MAX_EXP   ceil (log10( (double) _EXPBASE) * (double) DMAXEXP)
#define MIN_EXP   ceil (log10( (double) _EXPBASE) * (double) DMINEXP)

#define NUM_KEYWORDS          23  /* number of keywords in the keyword tables*/

static int LINECOUNT = 0;         /* number of lines seen in input file so
				     far */

/* table of keywords used in EMITID */
char *KEYWORD_TABLE[] = 
           {"AND","ARRAY","BEGIN","DIV","DO","ELSE","END",
	    "FORWARD","FUNCTION","IF","INTEGER","NOT",
	    "OF","OR","PROCEDURE","PROGRAM","READ","REAL",
	    "THEN","VAR","WHILE","WRITE","WRITELN"};

/* table of keyword's codes used in EMITID */
 int KEYWORD_CODE_TABLE[] =
           {AND_1,ARRAY_1,BEGIN_1,DIV_1,DO_1,ELSE_1,END_1,
	    FORWARD_1,FUNCTION_1,IF_1,INTEGER_1,NOT_1,
	    OF_1,OR_1,PROCEDURE_1,PROGRAM_1,READ_1,REAL_1,
	    THEN_1,VAR_1,WHILE_1,WRITE_1,WRITELN_1};

/* File I/O Variables and Functions in file stringIO.c*/
extern char CH;
extern char *LEXEME;

extern char *FRONT,*BACK;
extern unsigned LEX_LEN;

extern void FLUSH(void);
extern void GETCHR(void);
extern void BACKUP(void);

/* FORWARD function definitions */

static void SIGNEDEXPONENTQ(void);
static void EXPONENT(void);
static void FRACTION(void);
static void SIGNREALEXPONENTQ(void);

 
static int LETTER(int SYMB)
/* returns 0 if SYMB is not a letter, otherwise it returns the uppercase */
/* verison of the letter */
{
  if (isupper(SYMB))
    return SYMB;
  else if (islower(SYMB))
    return toupper(SYMB);
  else 
    return 0;
}


static int NONELETTER(int SYMB)
/* returns 0 if SYMB is not a letter or SYMB = 'E', otherwise it returns the */
/* uppercase verison of the letter: NOT STRONGLY TYPED */
{
  if (SYMB == 'E') return 0;
  else return LETTER(SYMB);
}


static int DIGIT(int SYMB)
/* returns 0 if SYMB is not a digit, otherwise it returns the CHARACTER for */
/* the digit */
{
  if (isdigit(SYMB))
    return SYMB;
  else
    return 0;
}


static int LETTERORDIGIT(int SYMB)

/* returns the character representing SYMB if SYMB is a letter or digit, */
/* otherwise it returns 0: NOT STRONGLY TYPED */
{
  char TEMP_SYMB;

  if (TEMP_SYMB = LETTER(SYMB))
    return TEMP_SYMB;
  else
    return DIGIT(SYMB);
}


static void EMITID(char LEXEME[])
/* prints uppercase version of LEXEME if it is not a keyword, otherwise it
   emits the code for the keyword in LEXEME */
{
  int I,       /* counter variable */
    CMP;       /* stores the result of a string comparison:
		  0 str1 = str2, 1 str1 > str2, -1 str1 < str2 */
  char CHR;

/* binary search variables */
  int FRNT = 0,               /* FRoNT of table searching for keyword */
    BCK,                      /* BaCK of table searching */
    MDDL,                     /* (FRNT+BCK)/2 (ie MiDDLe of table) */
    CODE;                     /* CODE of keyword at middle of table */

  BCK  = NUM_KEYWORDS - 1;
/* Convert all the lowercase letters to uppercase */
  for (I = 0; LEXEME[I] != 0; I++)
      if (! (LEXEME[I] = LETTERORDIGIT(CHR = LEXEME[I])))
	(void) printf("%d ERROR: EMITID called on a non-identifier: %d\n",
	       PERROR_1,CHR);

/* binary search */
  while (FRNT <= BCK) {
    MDDL = (FRNT + BCK) / 2;

    if (!(CMP = strcmp(KEYWORD_TABLE[MDDL],LEXEME))) {

/* found a keyword */
      CODE = KEYWORD_CODE_TABLE[MDDL];
      if (CODE == AND_1 || CODE == DIV_1)
	(void) printf("%d %d\n",MULOP_1,CODE);
      else
	if (CODE == OR_1)
	  (void) printf("%d %d\n",ADDOP_1,CODE);
	else
	  (void) printf("%d\n",CODE);
      BCK = NOT_FOUND_1;
    }
    else
/* have not found a keyword, but continue looking in appropriate half of the 
   table*/
      if (CMP > 0)
	BCK = MDDL - 1;
      else
	FRNT = MDDL + 1;
  }
/* not a keyword */
  if (BCK != NOT_FOUND_1)
    (void) printf("%d %s\n",ID_1,LEXEME);
}


static void EMITINTEGER(char LEX[])
/* print the integer representation of LEX (a string) as long as there is no
   overflow */
{
  int I,CONVERT = 0; /* I is a counter, CONVERT is the converted integer of
			the string */
  int FLAG = 1;     /* overflow FLAG: 1 = no overflow, 0 = overflow */

/* convert string and check for overflow */
  for (I = 0; (LEX[I] != 0 && FLAG); I ++)
    if (CONVERT < MAXINT/10)    /* if this true will not overflow with next
				   digit */
      CONVERT = 10*CONVERT + (LEX[I]-'0'); 
    else      /* may overflow */
      if (CONVERT == MAXINT/10 && (LEX[I]-'0') <= (MAXINT-(MAXINT/10)*10))
	      CONVERT = 10*CONVERT + (LEX[I]-'0'); /* didn't overflow */
      else {
	(void) printf("%d %d\n",INTEGER_NUMBER_1,ILLEGAL_1); /* did overflow */
	(void) printf("%d %s is a number that is too large\n",ERRORMSG_1,LEX);
	FLAG = 0;
	}

  if (FLAG)
    (void) printf("%d %d %d\n",INTEGER_NUMBER_1,VALID_1,CONVERT);
}


static void EMITREAL(char LEX[])
/* print the real representation of LEX (a string) as long as there is no
   overflow/underflow */
{
  int I,                  /* a counter variable */
    E_or_END,             /* a pointer to E in LEX or the END of LEX */
    EXP,                  /* EXPonent of LEX */
    I_POS_EXP = -1,       /* implicite positive Exponent (example:
			     1000 has an imp. pos. Exp. of 4 [1E4]) */
    E_EXP = 0,            /* explicite exponent (example:
			     1E4 has an expl. exp. of 4) */
    I_NEG_EXP = 0,        /* implicite negitive exponent (example:
			     .0001 had an imp. neg. exp. of -4 (1E-4)) */
    T_EXP = 1,            /* used to store parts of EXP do to zeros: if
			     there are leading zeros do not want to
			     include the in EXP, so store in T_EXP until
			     sure not leading zeros */
    PLUS_or_MINUS = 1;   /* is E_EXP positive(1) or negative (-1) */

  int FLAG = 1,           /* OVER/UNDER flow flag */
    T_FLAG = 1,           /* Temporary version of FLAG */
    ZERO_FLAG = 1,        /* flag for working on a string of leading zeros */
    DOT_FLAG = 0;         /* flag for if seen a . in LEX yet (and leading 
			     zeros) */

  double CONVERT,         /* LEX's real number representation */
    MANTISSA = 0.0,       /* LEX's mantissa */
    T_MANT = 1.0,         /* T_MANT used to handle leading zeros */
    R_EXP = 1.0;          /* Real Exponent: R_EXP = 1E(EXP) */

/* find out where the 'E' in LEX is or where the END of LEX is */
/* also set ZERO flag if have leading zeros (ex: 0.0001) and compute the
   value of the implicite negative exponent (I_NEG_EXP) (ex: -4 for 0.0001) */

  for (E_or_END = 0; (LEX[E_or_END] != 'E' && LEX[E_or_END] != 0); E_or_END ++)
    {
/* set ZERO_FLAG */
      if (LEX[E_or_END] == '.') {
	if (ZERO_FLAG)
	  DOT_FLAG = 1;
	FLAG = 1;
      }
      else 
	if (LEX[E_or_END] != '0') {
	  ZERO_FLAG = 0;
	  if (! DOT_FLAG) I_NEG_EXP = 0;
	}
	else
/* set I_NEG_EXP */
	  if (DOT_FLAG) 
	    if (I_NEG_EXP > -MAXINT)
	      I_NEG_EXP --;
	    else {
	      FLAG = 0;
	      EXP = MIN_EXP - 1;
	    }
    }

  if (! DOT_FLAG) {
    I_NEG_EXP = 0; /* no inplicite negatice exp. */
    FLAG = 1;
    }

/* find the mantissa and convert to a real number and set I_POS_EXP to
   it's correct value */
  for (I = E_or_END-1; I >= 0 && FLAG; I --)
    if (LEX[I] == '.') { 
      I_POS_EXP = -1;
      T_EXP = 1;
      T_FLAG = 1;
    }
    else {
      if (LEX[I] != '0') {
	MANTISSA = MANTISSA / 10 * T_MANT + (LEX[I] - '0');
	if (T_FLAG == -1)
	  T_FLAG = 0;
	if (I_POS_EXP <= MAXINT - T_EXP)
	  I_POS_EXP += T_EXP; 
	else
	  T_FLAG = 0;
	T_MANT = 1.0;
	T_EXP = 1;
      }
      else { /* make sure not just processing leading zeros */
	T_MANT = T_MANT / 10.0;
	if (T_EXP < MAXINT)
	  T_EXP ++;
	else T_FLAG = -1;
      }
    }

  if (FLAG && !T_FLAG) {
    FLAG = 0;
    EXP = MAX_EXP + 1;
  }
/* find the EXPLICITE EXPONENT and check for over/under flow */
  if (LEX[E_or_END] !=0 && FLAG && MANTISSA != 0.0) {
    E_or_END ++;
    if (LEX[E_or_END] == '-') {
      E_or_END ++;
      PLUS_or_MINUS = -1;
    }
    else
      if (LEX[E_or_END] == '+')
	E_or_END ++;
    for (I = E_or_END; LEX[I] != 0 ; I ++) {
      if (E_EXP < MAXINT/10 ||
	  (E_EXP == MAXINT/10 && (LEX[I] - '0') <= MAXINT%10))
	  E_EXP = E_EXP * 10 + (LEX[I] - '0');
      else {
	FLAG = 0;
	if (PLUS_or_MINUS == -1)
	  EXP = MIN_EXP -1;
	else
	  EXP = MAX_EXP + 1;
      }
    }
  }

  if (FLAG && MANTISSA !=0)
    /* can add I_POS_EXP and I_NEG_EXP because sum of a positive and negative
       number can not over/under flow */
    if ( (I_POS_EXP + I_NEG_EXP)*PLUS_or_MINUS == -1) {
      if ((PLUS_or_MINUS*E_EXP + (I_POS_EXP + I_NEG_EXP) < MIN_EXP)) {
	FLAG = 0;
	EXP = MIN_EXP - 1;
      }
      if ((PLUS_or_MINUS*E_EXP + (I_POS_EXP + I_NEG_EXP)) > MAX_EXP) {
	FLAG = 0;  
	EXP = MAX_EXP + 1;
      }
    }
    else
      if (PLUS_or_MINUS == -1) {
	if (-1*E_EXP < MIN_EXP - (I_POS_EXP + I_NEG_EXP))
	  {
	    FLAG = 0;
	    EXP = MIN_EXP -1;
	  }
      }
      else
	if (E_EXP > MAX_EXP - (I_POS_EXP + I_NEG_EXP))
	  {
	    FLAG = 0;
	    EXP = MAX_EXP+1;
	  }

  if (FLAG&&MANTISSA !=0) {
    EXP = PLUS_or_MINUS*E_EXP + I_POS_EXP + I_NEG_EXP;
    R_EXP = pow(10.0,EXP*1.0); 
  }

/* One last overflow check */
  if (EXP == MAX_EXP) 
    if (MANTISSA > (MAXDOUBLE / R_EXP)) 
      FLAG = 0; /* overflow */

  if (FLAG) { /* print convert */
    CONVERT = R_EXP * MANTISSA;
    (void) printf("%d %d %g\n",REAL_NUMBER_1,VALID_1,CONVERT);
  }
  else { /* over/underflow */
    (void) printf("%d %d\n",REAL_NUMBER_1,ILLEGAL_1);
    if (EXP < MIN_EXP)
      (void) printf("%d %s is a number that is too small\n",ERRORMSG_1,LEX);
    else
      (void) printf("%d %s is a number that is too large\n",ERRORMSG_1,LEX);
  }
}


/* The rest of the functions (and main) of this file is the encoded DFA
   from the Specification For The First Project */


static void INTEGERQ(void)
/* Seen: 1 digit */
{
  while (DIGIT(CH)) GETCHR(); /* get the largest string of just digits */
  if (CH == '.') { /* if first non digit is a . check if .. or a decimal
		      point of a real number or an error */
    GETCHR();
    if (DIGIT(CH)) FRACTION(); /* real number */
    else {
      if (CH != '.') /* .. */ {
	BACKUP();
	(void) printf("%d %c is an illegal integer delimiter\n",ERRORMSG_1,CH);
      }
      else
	BACKUP();
      BACKUP();
      EMITINTEGER(LEXEME);
      FLUSH();
    }
  }
  else
    if (CH == 'E') { /* may be a real number */
      GETCHR();
      if (DIGIT(CH)) EXPONENT(); /* is a real number */
      else 
	if (CH == '+' || CH == '-') SIGNEDEXPONENTQ(); /* may be real */
	else
	  {
	    BACKUP();
	    (void) printf("%d %c is an illegal integer delimiter\n",ERRORMSG_1,CH);
	    BACKUP();
	    EMITINTEGER(LEXEME);
	    FLUSH();
	  }
    }
    else {
      if (NONELETTER(CH)) /* if a NONeLETTER then it is an integer with an 
			     illegal delimeter */
	(void) printf("%d %c is an illegal integer delimiter\n",ERRORMSG_1,CH);
      BACKUP();
      EMITINTEGER(LEXEME);
      FLUSH();
    }
}


static void SIGNEDEXPONENTQ(void)
/* seen: (d+)E(+|-)
   check: if (d+)E(+|-)(exponent) */
{
  GETCHR();
  if (DIGIT(CH)) EXPONENT();
  else {
    BACKUP();
    BACKUP();
    (void) printf("%d %c is an illegal integer delimiter\n",ERRORMSG_1,CH);
    BACKUP();
    EMITINTEGER(LEXEME);
    FLUSH();
  }
}

static void EXPONENT(void)
/* looking for an exponent, get it */
{
  while (DIGIT(CH)) GETCHR();
  if (LETTER(CH) || CH == '.') 
    (void) printf("%d %c is an illegal real number delimiter\n",ERRORMSG_1,CH);
  BACKUP();
  EMITREAL(LEXEME);
  FLUSH();
}
 
static void FRACTION(void)
/* seen: (d+). 
   check if can finish a valid real number */
{
  while (DIGIT(CH)) GETCHR();
  if (CH == 'E') {
    GETCHR();
    if (DIGIT(CH)) EXPONENT();
    else 
      if (CH == '+' || CH == '-') SIGNREALEXPONENTQ();
      else
	{
	  BACKUP();
	  (void) printf("%d %c is an illegal real number delimiter\n",ERRORMSG_1,CH);
	  BACKUP();
	  EMITREAL(LEXEME);
	  FLUSH();
	}
  }
  else {
    if (NONELETTER(CH) || CH == '.') 
      (void) printf("%d %c is an illegal real number delimiter\n",ERRORMSG_1,CH);
    BACKUP();
    EMITREAL(LEXEME);
    FLUSH();
    }
}

static void SIGNREALEXPONENTQ(void)
/* seen: (d+).(D+)E(+|-)
   check: if (d+).(d+)E(+|-)(exponent) */

{
  GETCHR();
  if (DIGIT(CH)) EXPONENT();
  else {
    BACKUP();
    BACKUP();
    (void) printf("%d %c is an illegal real number delimiter\n",ERRORMSG_1,CH);
    BACKUP();
    EMITREAL(LEXEME);
    FLUSH();
  }
}

void main(void)
{
/* intialization */
  LEXEME = (char *) calloc(LEX_LEN,1);
  FLUSH();

/* ---------------------- START STATE ---------------------------- */
  GETCHR();
  if ((int) CH != EOF) 
    {
      LINECOUNT = 1;
      (void) printf("%d %d\n",LINE_1,LINECOUNT);
    }

/* ---------------------- NEWTOKEN STATE ---------------------------- */
/* encoding of the specification for the first project: see specs. */
  while ( (int) CH != EOF)
    {
      if (LETTER(CH)) {
	while (LETTERORDIGIT(CH)) GETCHR();
	BACKUP();
	EMITID(LEXEME);
	FLUSH();
      }
      else
	if (DIGIT(CH)) 
	  INTEGERQ();
	else
	  switch (CH) {
            case ',': (void) printf("%d\n",COMMA_1);
		    FLUSH();
		    break;  /*goto NEWTOKEN*/
		    
	    case ';': (void) printf("%d\n",SEMICOLON_1);
		      FLUSH();
		      break;  /*goto NEWTOKEN*/
		      
	    case '(': (void) printf("%d\n",LEFT_PAREN_1);
		      FLUSH();
		      break;  /*goto NEWTOKEN*/

	    case ')': (void) printf("%d\n",RIGHT_PAREN_1);
		      FLUSH();
		      break;  /*goto NEWTOKEN*/

            case '[': (void) printf("%d\n",LEFT_BRACKET_1);
		      FLUSH();
		      break;  /*goto NEWTOKEN*/

            case ']': (void) printf("%d\n",RIGHT_BRACKET_1);
		      FLUSH();
		      break;  /*goto NEWTOKEN*/

	    case '+':
            case '-': (void) printf("%d %d\n",ADDOP_1,CH);
		      FLUSH();
		      break;  /*goto NEWTOKEN*/

	    case '*':
	    case '/': (void) printf("%d %d\n",MULOP_1,CH);
		      FLUSH();
		      break;  /*goto NEWTOKEN*/

	    case '=': (void) printf("%d %d\n",RELOP_1,CH);
		      FLUSH();
		      break; /*goto NEWTOKEN*/

	    case ':': GETCHR();
		      if (CH == '=')
			(void) printf("%d\n",ASSIGN_1);
		      else {
			BACKUP();
			(void) printf("%d\n",COLON_1);
		      }
		      FLUSH();
		      break; /*goto NEWTOKEN*/

	    case '<': GETCHR();
		      if (CH == '>')
			(void) printf("%d %d\n",RELOP_1,NOTEQUAL_1);
		      else
			if (CH == '=')
			  (void) printf("%d %d\n",RELOP_1,LESSTHANOREQUAL_1);
			else {
			  BACKUP();
			  (void) printf("%d %d\n",RELOP_1,LESSTHAN_1);
			}
		      FLUSH();
		      break; /*goto NEWTOKEN*/

	    case '>': GETCHR();
		      if (CH == '=')
			(void) printf("%d %d\n",RELOP_1,GREATERTHANOREQUAL_1);
		      else {
			BACKUP();
			(void) printf("%d %d\n",RELOP_1,GREATERTHAN_1);
		      }
		      FLUSH();
		      break; /*goto NEWTOKEN*/

	    case '.': GETCHR();
		      if (CH == '.')
			(void) printf("%d\n",DOTDOT_1);
		      else {
			BACKUP();
			(void) printf("%d\n",DOT_1);
		      }
		      FLUSH();
		      break; /*goto NEWTOKEN*/

	    case '{': while (CH != '}' && (int) CH != EOF) {
	                if (CH == '\n') { 
			  LINECOUNT ++;
			  (void) printf("%d %d\n",LINE_1,LINECOUNT);
			}
			FLUSH();
	                GETCHR();
		      }

		      if ((int) CH == EOF) {
			(void) printf("%d end of file reached with unclosed comment\n"
			       , ERRORMSG_1);
			BACKUP();
		      }
		      FLUSH();
		      break;  /*goto NEWTOKEN*/

	    case ' ': 
	    case '\t':
	      FLUSH();
	      break;          /*goto NEWTOKEN*/
	    case '\n': 
	      LINECOUNT ++;
	      (void) printf("%d %d\n",LINE_1,LINECOUNT);
	      FLUSH();
	      break;          /*goto NEWTOKEN*/
	    default: if ((int) CH != EOF) 
	      (void) printf("%d encountered unknown character hex:%x\n",
			    ERRORMSG_1,CH);
	      FLUSH();
      } /* END of SWITCH (CH) */
      GETCHR();
    } /* END While ((int) CH != EOF) */
/* ---------------------- DONE ------------------------------- */
}
