/* this name of this file was pgm2.c */
/* This is a subset of mway_1.c */
/* This is a unified version for all best algorithms for mway partition */
/* This is the best mway.c up to Sept. 25 */
/* mway_2.c tried to use linked list representation but didn't really save */
 /* time. So this mway_1.c is better than mway_2.c. Mway_3.c will be a */
 /* subset of the good algorithms in this mway_1(); */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/types.h>
#include <sys/times.h>
#include <string.h>

#ifdef SOLARIS
#define srandom  srand
#define random   rand
#endif

#define MAXINT 1e7
#define	N_MAX	1201	/* Max node number plus 1 */
#define DEGREE_MAX 101	/* Max degree of each node plus 1 */
#define M_MAX 81       /* Max partition number plus 1 */
#define MaxProc 16      /* Max Parallel processors */


int w1[N_MAX];		/* Node weights */	
int w2[N_MAX][N_MAX];	/* Weighted adjacency matrix */	
int w3[N_MAX][N_MAX];	/* Transformed weighted adjacency matrix */	
int pi[N_MAX];		/* Allocation table for best partition */	
int v[N_MAX][DEGREE_MAX];
			/* [k][0]: number of neighbors for node k; */
                        /* [k][i]: node k's i-th neighbor */
int n;			/* Actual node number */
int m;			/* Actual partition number */
int ProcNum;            /* Number of processor available to use */
int choice;             /* Run which algorithm? */
FILE *fpin, *fpout, *fp_ins, *fp_plot1, *fp_plot2; 
char filein[15], fileout[15], fileplot1[15], fileplot2[15];
int seed=1;
int verbose=0;          /* 1: for more intermediate information; */
			/* 0: otherwise */
int graph_num;          /* Which graph is under processing now?*/
float r_factor;         /* Used for graph transformation */
int part[M_MAX][N_MAX]; /* part[k][0]: number of nodes in partition k */
                        /* part[k][i]: the i-th member of partion k */
float ModyFact;



int iter_num;           /* How many runs for an algorithm */
int mv_num, ex_num;     /* Total number for moves and exchanges performed */
int solutions[5001][3]; /* Store cost, w1, and w2 for all the iter_num runs */
float durations[5001];  /* Store all the iter_num running times */
int do_plotting1=1;     /* for different runs; 1: plot; 0: no plot */
int do_plotting2=0;     /* for different iterations; 1: plot; 0: no plot */
int plot_grain=500;     /* How many other iterations to plot a point */ 
char str[200];          /* Multi-purpose scratch */


struct tms tmsstart, tmsend;

/* Mark the start of the timer */
void time_start(void)
{ times(&tmsstart);}

/* Mark the stop of the timer */
void time_end(void)
{ times(&tmsend);}

/* Return CPU time in seconds between the start and stop of the timer */
float time_used(void)
{return ((float) (tmsend.tms_utime+tmsend.tms_cutime
        -tmsstart.tms_utime-tmsstart.tms_cutime) /60.0 );
}


/* Return random integer in [0, n-1] */
int random_int(int n)
{ return random()%n; /*(floor((double)((double) random()/dividend)*n));*/
}


double dividend = ((1 << 31) - 1);        /* 2^(31) -1 used by random01() */
/* Return random real number between 0 and 1 */
double random01(void)
{ return((double)random()/dividend);
}     


/* Print message s on screen and exit the program */
void error(char s[])
{ printf("%s\n", s);
  exit(1);
}


/* Dump string str to screan and fileout at the same time */
void dump(char str[])
{
  printf("%s", str);
  fprintf(fpout, "%s", str);
}

/* Transform w1[] and W2[][] into w3[][] */
void transform(void)
{ int R=0, i, j;

  for (i=1; i<n; i++)
    for (j=i+1; j<=n; j++) 
      R += w2[i][j];
  R++;
  R = ((float) R)*r_factor;
  for (i=1; i<=n; i++)
    for (j=1;j<=n; j++)
      w3[i][j] = w1[i]*w1[j]*R-w2[i][j];
}


/* Generate random initial solution */
void gen_ini_pi(int p[])
{ int i, t;

  for (i=1; i<=n; i++) p[i] = random_int(m)+1;
  for (i=1; i<=m; i++) part[i][0] = 0;
  for (i=1; i<=n; i++) {
    t = p[i];
    part[t][++part[t][0]] = i;
  }
}


/* Return the transformed cost of partition p */
int transformed_cost(int p[])
{ int i, j, t=0;
  
  for (i=1; i<n; i++)
    for (j=i+1; j<=n; j++) {
      if (p[i]==p[j]) continue;
      t += w3[i][j];
    }
  return t;
}

/* Return the original costs of partition p */
void  cost_w1_w2(int p[],int *cost_w1, int *cost_w2)
{ int node1, node2, part1, part2, c1[M_MAX], t1=0, t2=0;

  for (part1=1; part1<=m; part1++) c1[part1] = 0;
  for (node1=1; node1<=n; node1++) c1[p[node1]]+=w1[node1];
  for (part1=1; part1<m; part1++)
    for (part2=part1+1; part2<=m; part2++)
      t1 += abs(c1[part1]-c1[part2]);
  for (node1=1; node1<n; node1++)
    for (node2=node1+1; node2<=n; node2++) {
      if (p[node1]==p[node2]) continue;
      t2 += w2[node1][node2];
    }
  *cost_w1 = t1;
  *cost_w2 = t2;
}


/* Find the gain incurred by moving node_x to part_y */
int gain_move(int p[], int node_x, int part_y)
{ int from, l, node, gain=0;

  from = p[node_x];
  l = part[from][0];
  while (l> 0) {
    node = part[from][l]; 
    if (node==node_x) { l--; continue; }
    gain += w3[node_x][node];
    l--;
  }
  l = part[part_y][0];
  while (l > 0) {
    node = part[part_y][l]; 
    gain -= w3[node_x][node];
    l--;
  }
  return gain;
}



/* Move node_x to part_to */
void move_nasko(int p[],int node_x,int part_to)
{ int from, len, l;

  from = p[node_x];
  p[node_x] = part_to;
  len = part[from][0];
  for (l=1; l<=len && part[from][l]!=node_x; l++);
  if (l>len) printf("Move() failed: node_x=%d, to=%d\n", node_x, part_to);
  part[from][l] = part[from][len];
  part[from][0]--;
  l = ++part[part_to][0];
  part[part_to][l] = node_x;
}


/* Calculate gain for moving any node to any partitions */
void find_gain(int p[], int gains[][M_MAX])
{ int node, to;

  for (node=1; node<=n; node++) {
    for (to=1; to<=m; to++) {
      if (p[node]==to) 
	gains[node][to] = 0;
      else 
        gains[node][to] = gain_move(p, node, to);
    }
  }
}


/* Find non_fixed node_x and to_part to maximize gains[node_x][to_part] */
/* If there is more than one answers, a random one will be returned */
void find_max_gain(int gains[][M_MAX],int fixed[],int p[],
                   int *node_x,int *to_part)
{ int node, to, g, l, max_g=-MAXINT, max_node[N_MAX], max_to[N_MAX], max_last=0;

  for (node=1; node<=n; node++) {
    if (fixed[node]) continue;
    for (to=1; to<=m; to++) {
      g = gains[node][to];
      if (p[node]==to) continue;
      if (max_g<g) {
	max_g = g;
	max_node[1] = node;
	max_to[1] = to;
	max_last = 1;
      }
      else if (max_g==g) {
	max_last++;
	max_node[max_last] = node;
	max_to[max_last] = to;
      }
    }
  }
  l = random_int(max_last)+1;
  *node_x = max_node[l];
  *to_part = max_to[l];
}


/* Update gains[][] when node is moved from partition from to partition to */
/*  For all v in P[from],  g[v][to] -= 2*w3[v][node],         */
/*                         g[v][other] -= w3[v][node]         */
/*  For all v in P[to],    g[v][from] += 2*w3[v][node]        */
/*                         g[v][other] += w3[v][node]         */
/*  For all v in P[other], g[v][from] += w3[v][node]          */
/*                         g[v][to]   -= w3[v][node]          */
/*                         g[node][from] = - g[node][to]      */
/*                         g[node][other] -= g[node][to]      */
/*                         g[node][to] = 0   (last to modify) */
void update_gain(int g[][M_MAX],int p[],int fixed[],int node,int from,int to)
{ int l, v, to_part, temp;

  l = part[from][0];
  while (l) {
    v = part[from][l];
    l--;
    if (fixed[v] || v==node) continue;
    temp = w3[v][node];
    for (to_part=1; to_part<=m; to_part++) {
      if (to_part==from)
	g[v][from] = 0;
      else if (to_part==to) 
	g[v][to] -= (2*temp);
      else 
	g[v][to_part] -= temp;
    }
  }
  l = part[to][0];
  while (l) {
    v = part[to][l];
    l--;
    if (fixed[v]) continue;
    temp = w3[v][node];
    for (to_part=1; to_part<=m; to_part++) {
      if (to_part==to) 
	g[v][to] = 0;
      else if (to_part==from) 
	g[v][from] += (2*temp); 
      else 
	g[v][to_part] += temp; 
    }
  }
  for (v=1; v<=n; v++) {
    if (fixed[v] || p[v]==from ||p[v]==to) continue;
    g[v][from] += w3[v][node]; 
    g[v][to] -= w3[v][node];
  }
  /* Following only used by probes */
  if (fixed[node]) return;
  temp = g[node][to];
  for (to_part=1; to_part<=m; to_part++) {
    if (to_part==from)
      g[node][from] = -temp;
    else if (to_part!=to)
      g[node][to_part] -= temp;
  }
  g[node][to] = 0;
}


 
/* Global variables for mway_stochastic probe 1 */
int tot_iter_num, tolerance, no_imp, mody, StartMody, ex_max;


/* Modify the current solution to prepare to start a new probe */
void modify(int p[],int mody)
{ int i, k, p1, p2;

  for (k=1; k<=mody; k++) {
    p1=random_int(m)+1;
    while (part[p1][0]==0) p1=random_int(m)+1;
    p2=random_int(m)+1;
    while (p1==p2) p2=random_int(m)+1;
    i = random_int(part[p1][0])+1;
    move_nasko(p, i, p2);
  }
}


/* Find out moving node to which partition can maximize gain */
int find_move_fun(int node,int *to_part,int p[],int gains[][M_MAX])
{ int from, to, g, max_to[N_MAX], max_last=0, max_g=-MAXINT, l;

  from = p[node];
  for (to=1; to<=m; to++) {
    if (from==to) continue;
    g = gains[node][to];
    if (g>max_g) {
      max_g = g;
      max_to[1] = to;
      max_last = 1;
    }
    else if (g==max_g) 
      max_to[++max_last] = to;
  }
  l = random_int(max_last)+1;
  *to_part = max_to[l];
  return max_g;
}


/* Return the maximum gain induced by moving node1 from p[node1] to from */
 /* and (which) node2  from from to p[node1] */
int find_move_which_fun(int node1,int *node2,int from,int p[],
                        int gains[][M_MAX])
{ int to, node, g, max_g=-MAXINT, max_node[N_MAX], max_last=0, l;

  to = p[node1];
  l = part[from][0];
  while (l) {
    node = part[from][l];
    l--;
    g = gains[node1][from]+gains[node][to]+2*w3[node1][node];
    if (g>max_g) {
      max_g = g;
      max_node[1] = node;
      max_last = 1;
    }
    else if (g==max_g)
      max_node[++max_last] = node;
  }
  l = random_int(max_last)+1;
  *node2 = max_node[l];
  return max_g;
}


/* Stochastic Probe */
int mway_stochastic_probe_function(void)
{ int i,j, t,to_part, c_node=1, iter=0, no_impr_count=1, p[N_MAX], gain, r;
  int b_cost, cost, cost_w1, cost_w2, true_cost;
  int BestCost,StartCost,BestP[N_MAX],StartP[N_MAX];
  int BestPart[M_MAX][N_MAX],StartPart[M_MAX][N_MAX];
  int BestMvNum,BestExNum,TempMvNum,TempExNum,StartMvNum,StartExNum;
  int proc,PIter[MaxProc], MaxIterPerProb;
  int CurMody;
  int from, gain_1, gain_2, node_ex, last=0, most_negative_gain=0;
  int margin=tolerance, plot_count=0;
  int gains[N_MAX][M_MAX];   /* gains[i][k]: gain to move node i to */
			     /* partition k */
  int fixed[N_MAX];          /* dummy */

  for (i=1; i<=n; i++) fixed[i] = 0;
  mv_num=0, ex_num=0;
  BestMvNum=0, BestExNum=0;
  CurMody=StartMody;
  gen_ini_pi(p); 
  b_cost =cost = transformed_cost(p);
  BestCost=StartCost=cost;
  for (i=1;i<=m;i++) {
   t=part[i][0];
   for (j=0;j<=t;j++) StartPart[i][j]=BestPart[i][j]=part[i][j];
  }
  cost_w1_w2(p, &cost_w1, &cost_w2);
  sprintf(str,"Initial w1=%d, w2=%d, cost=%d\n", cost_w1, cost_w2, cost);
  dump(str);
  for (i=1; i<=n; i++) pi[i] =StartP[i]=BestP[i]= p[i];
  find_gain(p, gains);
  iter=0;
  while (iter<tot_iter_num) {
    if (verbose) 
      printf("b_cost=%3d, cost=%3d, c_node=%3d, to=%3d, gain=%3d, no_imp=%d, iter=%4d\n",
	   b_cost, BestCost, c_node, to_part, gain, no_impr_count, iter);
    if (do_plotting2 && ((iter % plot_grain)==1)) 
      fprintf(fp_plot2,"%5d %5d %5d %5d %5d %5d %4.3f\n", 
	      b_cost, BestCost, gain,iter,++plot_count,margin,0);
    if (b_cost<BestCost){
      b_cost = BestCost;
      for (i=1; i<=n; i++) pi[i] = BestP[i];
      last = iter;
    }
    for (i=1;i<=m;i++) {
      t=BestPart[i][0];
      for (j=0;j<=t;j++)  StartPart[i][j]=BestPart[i][j];
    }   
    for (i=1;i<=n;i++) StartP[i]=BestP[i];
   
    for (proc=1;proc<=ProcNum; proc++) {
     for (i=1; i<=n; i++) p[i] = StartP[i]; 
     for (i=1;i<=m;i++) {
      t=StartPart[i][0];
      for (j=0;j<=t;j++)  part[i][j]=StartPart[i][j];
     }
     TempMvNum=StartMvNum;
     TempExNum=StartExNum;
     modify(p, CurMody);
     cost = transformed_cost(p);
     find_gain(p, gains);
 /*    sprintf(str,"Solution modified: new cost = %d, last b_cost= %d\n",
              cost, b_cost); dump(str);   */
     no_impr_count = 0;
     MaxIterPerProb=0;
     PIter[proc]=0;
     while (no_impr_count< no_imp) {  
       PIter[proc]++;
       c_node++;
       if (c_node>n) c_node=1;
       from = p[c_node];
       gain_1 = find_move_fun(c_node, &to_part, p, gains);
       if (gain_1>0 || part[to_part][0]==0) gain_2 = -100;
       else {
         gain_2 = find_move_which_fun(c_node, &node_ex, to_part, p, gains);
       }
       if (gain_1>=gain_2) gain = gain_1;
       else                gain = gain_2; 
       if (gain<most_negative_gain) most_negative_gain = gain;
       r = -random_int(margin);
       if (gain>r) {
        update_gain(gains, p, fixed, c_node, from, to_part);
        move_nasko(p, c_node, to_part);
        if (gain==gain_2) {
	  update_gain(gains, p, fixed, node_ex, to_part, from);
	  move_nasko(p,node_ex,from); 
	  TempExNum++; 
        }
        else TempMvNum++;
        cost += gain;
       }
       if (gain>0) no_impr_count = 0;
       else        no_impr_count++;
     }
     if (BestCost<cost) {
         BestCost=cost;
         BestMvNum=TempMvNum;
         BestExNum=TempExNum;
         for (i=1;i<=n;i++) BestP[i]=p[i];
         for (i=1;i<=m;i++) {
           t=part[i][0];
           for (j=0;j<=t;j++) BestPart[i][j]=part[i][j];
         }

     }
     if (MaxIterPerProb<PIter[proc]) MaxIterPerProb = PIter[proc];
    }
    iter+=MaxIterPerProb;
    for (i=1; i<=n; i++) StartP[i]=BestP[i]; 
     for (i=1;i<=m;i++) {
      t=BestPart[i][0];
      for (j=0;j<=t;j++) StartPart[i][j]=BestPart[i][j];
     }
    StartMvNum=BestMvNum;
    StartExNum=BestExNum;
    CurMody=((float)CurMody)*ModyFact;
  }
  mv_num=BestMvNum, ex_num=BestExNum;
  true_cost = transformed_cost(pi);
  if (true_cost!=b_cost) printf("Failed\n");
  sprintf(str,"Total iteration number: %d\n",iter); dump(str);
  sprintf(str,"Last improvement at the %dth iteration\n", last); dump(str);
  sprintf(str,"Most negative gain = %d\n", most_negative_gain); dump(str);
  sprintf(str,"%d moves,  %d exchanges\n", mv_num, ex_num);   dump(str);
  return b_cost;
}


/* Return the best solution and its cpu time */
void best(int *cost,int *cost_w1,int *cost_w2,float *duration,float *ave_w1,
          float *ave_w2,float *ave_g_cost,float *ave_time)
{ int i, j, t=0;

  *ave_w1=*ave_w2=*ave_g_cost=*ave_time=0.0;
  for (i=1; i<=iter_num; i++) {
    if (solutions[i][0]>t) {
      t = solutions[i][0];
      j = i;
    }
    *ave_g_cost += solutions[i][0];
    *ave_w1 += solutions[i][1];
    *ave_w2 += solutions[i][2];
    *ave_time += durations[i];
  }
  *cost = solutions[j][0];
  *cost_w1 = solutions[j][1];
  *cost_w2 = solutions[j][2];
  *duration = durations[j];
  *ave_g_cost /= iter_num;
  *ave_w1 /= iter_num;
  *ave_w2 /= iter_num;
  *ave_time /= iter_num;
}


void dispatcher(void)
{ int number, i, j, k, dg, w, serial;
  int cost, cost_w1, cost_w2, sum1_cost=0, sum2_cost=0;
  float duration, sum_duration=0, ave_g_cost, ave_w1, ave_w2, ave_time;
  
  system("clear");
  printf("Enter the file name of the graphs > ");
  fscanf(fp_ins, "%s", filein);
  strcpy(fileout, filein);
  sprintf(str, ".ou%d", choice);
  strcat(fileout, str); 
  strcpy(fileplot1, filein);
  sprintf(str, ".1d%d", choice);
  strcat(fileplot1, str); 
  strcpy(fileplot2, filein);
  sprintf(str, ".2d%d", choice);
  strcat(fileplot2, str); 
  strcat(filein, ".in");
  fpin = fopen(filein, "r");
  fpout = fopen(fileout, "w");
  if (do_plotting1) { 
    fp_plot1 = fopen(fileplot1, "w");
    fclose(fp_plot1);
  }
  if (do_plotting2) {
    
    fp_plot2 = fopen(fileplot2, "w");
    fclose(fp_plot2);
  }
  printf("Current input file:  %s\n", filein);
  fprintf(fpout, "Current input file:  %s\n", filein); 
  printf("Partition number m > ");
  fscanf(fp_ins, "%d", &m);
  fprintf(fpout, "m = %d\n", m);
  printf("Algorithm iteration number > ");
  fscanf(fp_ins, "%d", &iter_num);
  printf("Input r_factor > ");
  fscanf(fp_ins, "%f",  &r_factor);
  fprintf(fpout, "r_factor=%f\n", r_factor);
    fprintf(fpout, "Current algorithm:   mway_stochastic_probe_function()\n");
    printf("Input Tot_iter_num, tolerance, no_imp, probe_mody, start_mody, mody_fact, proc#\n");
    fscanf(fp_ins, "%d %d %d %d %d %f %d", &tot_iter_num,  &tolerance, &no_imp,
	   &mody, &StartMody, &ModyFact, &ProcNum);
    fprintf(fpout, 
	    "tot_iter_num =%d, tolerance =%d, no_imp =%d, probe_mody =%d, start_mody =%d, mody_fact=%f, proc#=%d\n", 
	    tot_iter_num, tolerance, no_imp, mody,StartMody, ModyFact, ProcNum);
    fprintf(fpout,"______________________________________________________\n");
  fclose(fpout);
  fscanf(fpin, "%d", &number); 
  for (graph_num=1; graph_num<=number; graph_num++) {
    srandom(seed);
    fpout = fopen(fileout, "a");
    if (do_plotting2) fp_plot2 = fopen(fileplot2, "a");
    fscanf(fpin, "%d", &serial); 
    fscanf(fpin, "%d", &n);
    for (i=1; i<=n; i++) fscanf(fpin, "%d", &w1[i]);
    for (i=1; i<=n; i++)
      { for (j=1; j<=n; j++)  v[i][j] = w2[i][j] = w2[j][i] = 0;
	fscanf(fpin, "%d", &dg);
	v[i][0] = dg;
	for (k=1; k<= dg; k++) {
	  fscanf(fpin, "%d %d", &j, &w);
	  v[i][k] = j;
	  w2[i][j] = w2[j][i] = w;
	}
      }
    fprintf(fpout, "Serial number: %d\n", serial);
    fprintf(fpout, "n = %3d\n", n);     
    transform();
    for (i=1; i<=iter_num; i++) {
      srandom(i);
      if (do_plotting1) fp_plot1 = fopen(fileplot1, "a");
      time_start();
 	cost = mway_stochastic_probe_function();
      time_end();	
      duration = time_used()/ProcNum;
      cost_w1_w2(pi, &cost_w1, &cost_w2);
      solutions[i][0] = cost;
      solutions[i][1] = cost_w1;
      solutions[i][2] = cost_w2;
      durations[i] = duration;
      printf("Iter=%d, cost=%d, w1=%d, w2=%d, time=%f\n",
	     i, cost, cost_w1, cost_w2, duration);
      if (do_plotting1)
	fprintf(fp_plot1, "%5d %5d  %5.3f %5d\n",  cost_w1, cost_w2,
		duration, i);
      sprintf(str, "______________________________________________________\n");
      dump(str);
      if (do_plotting1) fclose(fp_plot1);
    }
    best(&cost, &cost_w1, &cost_w2, &duration, &ave_w1, &ave_w2, 
	 &ave_g_cost, &ave_time);
    sum_duration += duration;
    sum1_cost += cost_w1; sum2_cost += cost_w2;
    sprintf(str,"Graph %3d finished in %9.9f sec. Best:  W1 = %d, W2 = %d, cost = %d\n",  
	    graph_num, duration, cost_w1, cost_w2, cost);
    dump(str);
    sprintf(str,"Ave_W1 = %.3f, Ave_W2 = %.3f, Ave_cost = %.3f, Ave_time = %.3f\n",  
	    ave_w1, ave_w2, ave_g_cost, ave_time);
    dump(str);

    fprintf(fpout,"______________________________________________________\n");
    fclose(fpout);
    if (do_plotting2) fclose(fp_plot2);
  }
  ave_w1 = ((float) sum1_cost)/number;
  ave_w2 = ((float) sum2_cost)/number;
  ave_time = sum_duration/number;
  fpout = fopen(fileout, "a");
  fprintf(fpout, "******************************************************\n");
  fprintf(fpout,
	  "Average W1 = %5.3f, average W2 = %5.3f, average time = %9.9f sec.\n", 
	  ave_w1, ave_w2, ave_time);
  fclose(fpout);
  printf("******************************************************\n");
  printf(
	 "Average W1 = %5.3f, average W2 = %5.3f, average time = %9.9f sec.\n", 
	 ave_w1, ave_w2, ave_time);
  fclose(fpin);
}




void main(void)
{
  char  hit[5];

  seed = 1;
  fp_ins = stdin; /* fopen("instruct.txt", "r");   */
  system("clear");
  while (1) {
    printf("\n                      MENU\n\n\n");
    printf("   1: Run mway_stochastic_probe\n");
    printf("   9: Change verbose mode\n");
    printf("  10: Change plot mode\n");
    printf("   0: Exit\n");
    printf("\n\n   Verbose = %d   plot = %d\n", verbose, do_plotting2);
    printf("\n\n Your choice > ");
    fscanf(fp_ins, "%d", &choice);
 /*  system("mv -f *.ou* log");
      if (do_plotting1 || do_plotting2) system("mv -f *.?d? plot"); */

    switch (choice) {
    case 0:      exit(0);
    case 9: printf("1: more information; 0: otherwise > ");
      fscanf(fp_ins, "%d", &verbose);
      break;
    case 10: printf("1: plot for iterations; 0: no plot > ");
      fscanf(fp_ins, "%d", &do_plotting2);
      break;
    default: 
      dispatcher(); break;
    }
    printf("Type any key then RETURN to continue ... ");
    scanf("%s", hit);
    system("clear");
  }
}
