/****************************************************************************/
/*                                                                          */
/*                        Dermining Graph connexity                         */
/*                                                                          */
/****************************************************************************/

#include "div.h"
/* 
 Values for lien : 
   0 -> no link
   1 -> simple ascending link

   3 -> same level link { internal node }
   4 -> same level link { weak connexity }
   5 -> same level link { strong connexity }

  10 -> decseending connecting link
  11 -> ascending connecting link
  12 -> link along 3i+1 irregular triangle 
*/


/* Function Strong_connexity() */ 

void Strong_connexity(triangulation trngl,int **lien,int *color,
                      int *trn_color,int *connect,int *voisins,
                      int *frere,int *soeur,int *nc1)
{

  int PTS=trngl.PTS, *nb_mat=trngl.nb_mat;
  int **point=trngl.point, **triangle=trngl.triangle; 

  int i,j,k,c1,c2;

  for(i=1;i<=PTS;i++) {
			connect[i]=i;
			voisins[i]=0;
		      }

  /********************* Finding locally strong links ***********************/

  trn_color[0]=2;
  for(i=1;i<=PTS;i++)
    for(j=0;j<nb_mat[i];j++)
      if(trn_color[triangle[i][j]]%3==2)
	if(color[i]==color[point[i][j]])
	if(trn_color[triangle[i][(j-1+nb_mat[i])%nb_mat[i]]]%3!=2)
	  {
	    k=point[i][j];
	    lien[i][j]=5;
	    voisins[i]++;
	    frere[i]=j;
	    soeur[k]=i;
	    for(c1=i;connect[c1]!=c1;c1=connect[c1]);
	    for(c2=k;connect[c2]!=c2;c2=connect[c2]);
	    if(c2<c1) connect[c1]=c2; else connect[c2]=c1;
	 }

  /**************** Finding general strong connexity & statistics ***********/

  for(i=1;i<=PTS;i++) connect[i]=connect[connect[i]];
  for(i=1;i<=PTS;i++)
    if(connect[i]==i)
      {
	k=0;
	for(j=1;j<=PTS;j++)
	  if(connect[j]==i) k++;
	nc1[i]=k;
      };
}

/* Function Super_Srong_Connexity() */

void Super_Srong_Connexity(triangulation trngl,int nocolor,int *voisins,
                           int *frere,int *trn_color,int *con_trn0)
{
  int PTS=trngl.PTS, TRN=trngl.TRN, *nb_mat=trngl.nb_mat;
  int **point=trngl.point, **triangle=trngl.triangle; 
 
  int i,j,k,c1,c2,jfi,kfj;

  /************* Links of con_trn0 are on triangles, not vertex *************/

  /************* Linking locally simple vertex ( non-multiple ) *************/

  for(i=1;i<=TRN;i++) con_trn0[i]=i;
  for(i=1;i<=PTS;i++)
    if(voisins[i]==1)
      if(voisins[point[i][frere[i]]]==1)
	{
	  jfi=frere[i];
	  j=point[i][jfi];
	  kfj=frere[j];
	  k=point[j][kfj];
	  if(triangle[i][jfi]!=point[j][kfj])
	    {
	      for(c1=triangle[i][jfi];con_trn0[c1]!=c1;c1=con_trn0[c1]);
	      for(c2=triangle[j][kfj];con_trn0[c2]!=c2;c2=con_trn0[c2]);
	      if(c2<c1) con_trn0[c1]=c2; else con_trn0[c2]=c1;
	    }
	}
  for(i=1;i<=TRN;i++) con_trn0[i]=con_trn0[con_trn0[i]];

  /******************** Linking internally triangles ************************/

  for(i=1;i<=PTS;i++) for(j=0;j<nb_mat[i];j++)
    if(trn_color[triangle[i][j]]/3==trn_color[triangle[i][(j+1)%nb_mat[i]]]/3)
      {
	for(c1=triangle[i][j];con_trn0[c1]!=c1;c1=con_trn0[c1]);
	for(c2=triangle[i][(j+1)%nb_mat[i]];con_trn0[c2]!=c2;c2=con_trn0[c2]);
	if(c2<c1) con_trn0[c1]=c2; else con_trn0[c2]=c1;
      }
  for(i=1;i<=TRN;i++) con_trn0[i]=con_trn0[con_trn0[i]];
}

/* Function Strong_Connexity_Numbering() */

void Strong_Connexity_Numbering(triangulation trngl,int **lien,int *frere,
     int *soeur,int *connect,int *nc1,int *con_trn0, int *voisins,int *c_order,
     int *connect0,int *enter_node,int *exit_node,int *trn_color)

{
  int PTS=trngl.PTS, *nb_mat=trngl.nb_mat;
  int **point=trngl.point, **triangle=trngl.triangle;
 
  int i,j,k,l,m;
  int pile1[Max_pts],pile2[Max_pts],hpile,last_node;

  /************************** Looking each strong component *****************/

  for(i=1;i<=PTS;i++)
    if(connect[i]==i&&nc1[i]>2)
      {
	k=i;
	last_node=soeur[i];
	hpile=0;
	pile1[0]=0;
	for(j=1;j<=nc1[i];j++)
	  {
	    if(voisins[k]>1)
	      {
		/************ Multiple vertex ***********/
		m=nb_mat[k];
		if(pile1[hpile]!=k)
		  {
		    for(l=0;l<m&&point[k][l]!=last_node;l++);
		    pile1[++hpile]=k;
		    pile2[hpile]=l;
		    soeur[k]=last_node;
		    connect0[k]=con_trn0[triangle[k][l]];
		  }
		else exit_node[con_trn0[triangle[k][(l+1)%m]]]=j;
		
		for(l=(pile2[hpile]-1+m)%m;lien[k][l]!=5;l=(l-1+m)%m);
		pile2[hpile]=l;
		if(con_trn0[triangle[k][(l-1+m)%m]]==connect0[k])
		  {
		    c_order[k]=j;
		    frere[k]=l;
		    hpile--;
		  }
		else enter_node[con_trn0[triangle[k][l]]]=j--; 
	      }
	    else
	      {
		/************ Other ones ... ***********/
		c_order[k]=j;
		l=frere[k];
		connect0[k]=con_trn0[triangle[k][l]];
	      }
	    last_node=k;
	    k=point[k][l];
	  }
      }
}

/* Function Weak_Connexity() */

void Weak_Connexity(triangulation trngl,int **lien,int *connect,int *color,
                    int *connect2,int *link,int *sons,int *nc2)
{
  int PTS=trngl.PTS, *nb_mat=trngl.nb_mat;
  int **point=trngl.point;
 
  int i,j,k,c,c1,c2;

  for(i=1;i<=PTS;i++) connect2[i]=connect[i];

  /********** Checking each mono-color link and comparing components ********/

  for(i=1;i<=PTS;i++)
    for(j=0;j<nb_mat[i];j++)
      {
	k=point[i][j];
	if(color[i]==color[k]&&connect[i]!=connect[k])
	  {
	    lien[i][j]=4;
	    for(c1=i;c1!=connect2[c1];c1=connect2[c1]);
	    for(c2=k;c2!=connect2[c2];c2=connect2[c2]);
	    if(c1>c2) connect2[c1]=c2; else connect2[c2]=c1;
	  }
      }
  for(i=1;i<=PTS;i++) connect2[i]=connect2[connect2[i]];

  /********** Computing usefull statistics on weak components ***************/

  for(i=1;i<=PTS;i++) link[i]=0;
  for(i=1;i<=PTS;i++)
    for(j=0;j<nb_mat[i];j++)
      {
	k=point[i][j];
	if(color[i]!=color[k])
	  if(color[i]>color[k]) link[i]=k;
			   else link[k]=i;
      }
  for(i=1;i<=PTS;i++) sons[i]=0;
  for(i=1;i<=PTS;i++) nc2[i]=0;
  for(i=1;i<=PTS;i++)
    if(connect2[i]==i)
      {
	k=0;
	for(j=1;j<=PTS;j++)
	  if(connect2[j]==i) k++;
	for(c=i;c!=0;c=connect2[link[c]]) nc2[c]+=k;
	for(c=connect2[link[i]];c!=0&&sons[c]!=0;c=connect2[link[c]])
	   sons[c]++;
	if(c!=0) sons[c]=1;
      }
}

