/****************************************************************************/
/*                                                                          */
/*                     Graph basic computations                             */
/*                                                                          */
/****************************************************************************/

#include "div.h"

void Load_Mesh(triangulation *trngl)
{
  int PTS=0,TRN=0;
  int *nb_mat=trngl->nb_mat;
  int **point=trngl->point, **triangle1=trngl->triangle;
  int *ptx=trngl->ptx, *pty=trngl->pty,(*p)[Max_trn]=trngl->p;
  
  FILE *f;
  char s[100];
  int *triangle2[Max_pts];
  int i,j,k,l,c,pt;

/*********************** read data from file *******************************/

  f = fopen(SOURCE_NAME,"r");
  while(fgets(s,100,f))
    { 
      if(!sscanf(s,"T %i %i %i",&i,&j,&k))
	{
	  sscanf(s,"N%i %i %i",&i,&j,&k);
	  ptx[++(PTS)]=j;
	  pty[PTS]=k;
	}
      else
	{
	  p[0][++TRN]=i; p[1][TRN]=j; p[2][TRN]=k;
	  nb_mat[i]++;    nb_mat[j]++;  nb_mat[k]++;
	}
    }

  trngl->PTS=PTS;
  trngl->TRN=TRN;

/************** allocate & initialize fondamental variables ******************/

  for(i=1;i<=PTS;i++)
    {
      point[i]=(int *)malloc((nb_mat[i]+1)*sizeof(int));
      triangle1[i]=(int *)malloc((nb_mat[i]+1)*sizeof(int));
      triangle2[i]=(int *)malloc((nb_mat[i]+1)*sizeof(int));
      for(j=0;j<nb_mat[i];j++) triangle2[i][j]=0;
      nb_mat[i]=0;
    }
  for(i=1;i<=TRN;i++)
    for(j=0;j<3;j++)
      {
	pt=p[j][i];
	point[pt][nb_mat[pt]]=p[(j+1)%3][i];
	triangle1[pt][nb_mat[pt]++]=i;
      }


/************************** start borders & link ***************************/
  
  for(i=1;i<=PTS;i++)
    for(j=0;j<nb_mat[i];j++)
      {
	k=point[i][j];
	for(l=0;l<nb_mat[k]&&point[k][l]!=i;l++);
	if(l!=nb_mat[k]) triangle2[i][j]=triangle1[k][l];
		    else {
			   point[k][nb_mat[k]]=i;
			   triangle1[k][nb_mat[k]]=0;
			   triangle2[k][nb_mat[k]++]=triangle1[i][j];

			   c=point[i][0];
			   point[i][0]=point[i][j];
			   point[i][j]=c;

			   c=triangle1[i][0];
			   triangle1[i][0]=triangle1[i][j];
			   triangle1[i][j]=c;

			   triangle2[i][j]=triangle2[i][0];
			   triangle2[i][0]=0;
			 }
      }


/************************ reorder point description *************************/

  for(i=1;i<=PTS;i++)
    for(j=1;j<nb_mat[i]-1;j++)
      {
	k=triangle1[i][j-1];
	for(l=j;l<nb_mat[i]&&triangle2[i][l]!=k;l++);
	c=point[i][j];     point[i][j]=point[i][l];         point[i][l]=c;
	c=triangle1[i][j]; triangle1[i][j]=triangle1[i][l]; triangle1[i][l]=c;
	c=triangle2[i][j]; triangle2[i][j]=triangle2[i][l]; triangle2[i][l]=c;
      }

/*************************** freing of triangle2 ****************************/

  for(i=1;i<=PTS;i++) free(triangle2[i]);
}

void Set_Colors(triangulation trngl,int *nocolor,int *color,int *trn_color)

/*int PTS, TRN, *nb_mat, **point, **triangle, *ptx, *pty, p[3][Max_trn];*/

{
  int PTS=trngl.PTS, TRN=trngl.TRN, *nb_mat=trngl.nb_mat;
  int **point=trngl.point, **triangle=trngl.triangle; 
  int (*p)[Max_trn]=trngl.p;
  
  int i,j,k,nbcolor;

  /************************** Vertex colors *********************************/

  for(i=1;i<=PTS;i++) color[i]=0;
  nbcolor=0;
  for(i=1;i<=PTS;i++)
    if(triangle[i][nb_mat[i]-1]==0)
      {
	color[i]=1;
	nbcolor++;
          
        j=point[i][0];
      }
  for(*nocolor=1;nbcolor<PTS;(*nocolor)++)
    for(i=1;i<=PTS;i++)
      for(j=0;j<nb_mat[i];j++)
       {
	 k=point[i][j];
	 if(color[i]==0&&color[k]==*nocolor)
	   {
	     color[i]=*nocolor+1;
	     nbcolor++;
	   }
       }

  /************************** Triangle colors *******************************/

  for(i=1;i<=TRN;i++)
    trn_color[i]=color[p[0][i]]+color[p[1][i]]+color[p[2][i]];
}

/* Function Pre_Analyse() */

void Pre_Analyse(triangulation *trngl,int **lien,int *color,int *trn_color)
{
  int PTS=trngl->PTS, TRN=trngl->TRN, *nb_mat=trngl->nb_mat, **point=trngl->point;
  int (*p)[Max_trn]=trngl->p;
  int i,j,c0,c1,c2,colour;

  /************************** lien initialization ***************************/

  for(i=1;i<=PTS;i++) 
    {
      lien[i]=(int *)malloc(nb_mat[i]*sizeof(int));
      for(j=0;j<nb_mat[i];j++) if(color[i]==color[point[i][j]]) lien[i][j]=3;
			       else if(color[i]>color[point[i][j]]) 
					lien[i][j]=1;
				    else
					lien[i][j]=2;
    }

  /************************* triangles redirecting **************************/

  for(i=1;i<=TRN;i++)
    switch(trn_color[i]%3)
      {
	case 1:
	  colour=trn_color[i]/3;
	  for(j=0;j<3&&color[p[j][i]]!=colour+1;j++);
	  c0=p[j][i];
	  c1=p[(j+1)%3][i];
	  c2=p[(j+2)%3][i];
	  p[0][i]=c0;
	  p[1][i]=c1;
	  p[2][i]=c2;
	  break;
	case 2:
	  colour=trn_color[i]/3;
	  for(j=0;j<3&&color[p[j][i]]!=colour;j++);
	  c0=p[j][i];
	  c1=p[(j+1)%3][i];
	  c2=p[(j+2)%3][i];
	  p[0][i]=c0;
	  p[1][i]=c1;
	  p[2][i]=c2;
	  break;
      }
}
