#
#include "defs.h"

extern void PQinitialize(void);
extern struct Site *readone(void);
extern struct Site *nextone(void);
extern void out_site(struct Site *s);
extern void ELinitialize(void);
extern int PQempty(void);
extern struct Point PQ_min(void);
extern struct Halfedge *ELleftbnd(struct Point *p);
extern struct Halfedge *ELright(struct Halfedge *he);
extern struct Site *rightreg(struct Halfedge *he);
extern struct Edge *bisect(struct Site *s1,struct Site *s2);
extern struct Halfedge *HEcreate(struct Edge *e,int pm);
extern void ELinsert(struct Halfedge *lb,struct Halfedge *new);
extern struct Site *intersect(struct Halfedge *el1,struct Halfedge *el2);
extern void PQdelete(struct Halfedge *he);
extern void PQinsert(struct Halfedge *he, struct Site *v, float offset);
extern float dist(struct Site *s,struct Site *t);
extern struct Halfedge *PQextractmin(void);
extern struct Halfedge *ELleft(struct Halfedge *he);
extern struct Site *leftreg(struct Halfedge *he);
extern void out_triple(struct Site *s1,struct Site *s2,struct Site *s3);
extern void makevertex(struct Site *v);
extern void endpoint(struct Edge *e, int lr, struct Site *s);
extern void ELdelete(struct Halfedge *he);
extern void deref(struct Site *v);
extern void out_ep(struct Edge *e);

/* implicit parameters: nsites, sqrt_nsites, xmin, xmax, ymin, ymax,
   deltax, deltay (can all be estimates).
   Performance suffers if they are wrong; better to make nsites,
   deltax, and deltay too big than too small.  (?) */

void voronoi(int triangulate,int Sorted)
{
struct Site *newsite, *bot, *top, *temp, *p;
struct Site *v;
struct Point newintstar;
int pm;
struct Halfedge *lbnd, *rbnd, *llbnd, *rrbnd, *bisector;
struct Edge *e;


PQinitialize();
/* bottomsite = (*nextsite)(); */
if (Sorted)
  bottomsite = readone();
else
  bottomsite = nextone();
out_site(bottomsite);
ELinitialize();

/* newsite = (*nextsite)(); */
if (Sorted)
  newsite = readone(); 
else
  newsite = nextone(); 
while(1)
{
	if(!PQempty()) newintstar = PQ_min();

	if (newsite != (struct Site *)NULL 
	   && (PQempty() 
		 || newsite -> coord.y < newintstar.y
	 	 || (newsite->coord.y == newintstar.y 
		     && newsite->coord.x < newintstar.x)))
	{/* new site is smallest */
		out_site(newsite);
		lbnd = ELleftbnd(&(newsite->coord));
		rbnd = ELright(lbnd);
		bot = rightreg(lbnd);
		e = bisect(bot, newsite);
		bisector = HEcreate(e, le);
		ELinsert(lbnd, bisector);
		if ((p = intersect(lbnd, bisector)) != (struct Site *) NULL) 
		{	PQdelete(lbnd);
			PQinsert(lbnd, p, dist(p,newsite));
		};
		lbnd = bisector;
		bisector = HEcreate(e, re);
		ELinsert(lbnd, bisector);
		if ((p = intersect(bisector, rbnd)) != (struct Site *) NULL)
		{	PQinsert(bisector, p, dist(p,newsite));	
		};
		/* newsite = (*nextsite)();  */
		if (Sorted)
		  newsite = readone(); 
		else
		  newsite = nextone(); 
	}
	else if (!PQempty()) 
	/* intersection is smallest */
	{	lbnd = PQextractmin();
		llbnd = ELleft(lbnd);
		rbnd = ELright(lbnd);
		rrbnd = ELright(rbnd);
		bot = leftreg(lbnd);
		top = rightreg(rbnd);
		out_triple(bot, top, rightreg(lbnd));
		v = lbnd->vertex;
		makevertex(v);
		endpoint(lbnd->ELedge,lbnd->ELpm,v);
		endpoint(rbnd->ELedge,rbnd->ELpm,v);
		ELdelete(lbnd); 
		PQdelete(rbnd);
		ELdelete(rbnd); 
		pm = le;
		if (bot->coord.y > top->coord.y)
		{	temp = bot; bot = top; top = temp; pm = re;}
		e = bisect(bot, top);
		bisector = HEcreate(e, pm);
		ELinsert(llbnd, bisector);
		endpoint(e, re-pm, v);
		deref(v);
		if((p = intersect(llbnd, bisector)) != (struct Site *) NULL)
		{	PQdelete(llbnd);
			PQinsert(llbnd, p, dist(p,bot));
		};
		if ((p = intersect(bisector, rrbnd)) != (struct Site *) NULL)
		{	PQinsert(bisector, p, dist(p,bot));
		};
	}
	else break;
};

for(lbnd=ELright(ELleftend); lbnd != ELrightend; lbnd=ELright(lbnd))
	{	e = lbnd -> ELedge;
		out_ep(e);
	};
}
