#include <curses.h>
#include <signal.h>
#include <unistd.h>
#include "constants.h"
#include "types.h"
#include "externs.h"

extern int total_winner;
extern int moria_flag;
extern int search_flag;

#ifdef USG
/* no local special characters */
#else
extern struct ltchars save_special_chars;
#endif

int error_sig, error_code;
int (*core_dump)();
#ifdef USG
/* no suspend signal */
#else
int (*suspend_handler)();
#endif

/* This signal package was brought to you by           -JEW-  */

/* (*signal())(); */
#ifdef USG
void signal_save_core(int sig);
void signal_save_no_core(int sig);
void signal_ask_quit(int sig);
#else
void signal_save_core(int sig, int code, struct sigcontext *scp);
void signal_save_no_core(int sig, int code, struct sigcontext *scp);
void signal_ask_quit(int sig, int code, struct sigcontext *scp);
#endif

extern void prt(char *str_buff, int row, int col);
extern void save_char(int exit, int no_ask);
extern void exit_game(void);
extern int get_com(char *prompt, char *command);
extern void upon_death(void);
extern void search_off(void);
extern void rest_off(void);
extern void erase_line(int row, int col);

void init_signals(void)
{
  (void) signal(SIGHUP, signal_save_no_core);
  (void) signal(SIGINT, signal_ask_quit);
  core_dump = signal(SIGQUIT, signal_save_core);
  (void) signal(SIGILL, signal_save_core);
  (void) signal(SIGTRAP, signal_save_core);
  (void) signal(SIGIOT, signal_save_core);
  (void) signal(SIGEMT, signal_save_core);
  (void) signal(SIGFPE, signal_save_core);
  (void) signal(SIGKILL, signal_save_core);
  (void) signal(SIGBUS, signal_save_core);
  (void) signal(SIGSEGV, signal_save_core);
  (void) signal(SIGSYS, signal_save_core);
  (void) signal(SIGTERM, signal_save_core);
}

/*ARGSUSED*/
#ifdef USG
void signal_save_core(int sig)
{
  error_sig = sig;
  error_code = 0;
  prt("OH NO!!!!!!!!!!  Attempting panic save.", 23, 0);
  save_char(FALSE, FALSE);
  (void) signal(SIGQUIT, SIG_DFL);
  /* restore terminal settings */
#ifndef BUGGY_CURSES
  nl();
#endif
#ifdef ultrix
  nocrmode();
#else
  nocrmode();
#endif
  echo();
  resetterm1();
  /* restore the saved values of the local special chars */
  /* no local special characters */
  /* allow QUIT signal */
  /* nothing needs to be done here */
  (void) kill(getpid(), 3);
  exit_game();
}
#else
void signal_save_core(int sig, int code, struct sigcontext *scp)
{
  error_sig = sig;
  error_code = code;
  prt("OH NO!!!!!!!!!!  Attempting panic save.", 23, 0);
  save_char(FALSE, FALSE);
  (void) signal(SIGQUIT, SIG_DFL);
  /* restore terminal settings */
#ifndef BUGGY_CURSES
  nl();
#endif
#ifdef ultrix
  nocrmode();
#else
  nocrmode();
#endif
  echo();
  /* restore the saved values of the local special chars */
  (void) ioctl(0, TIOCSLTC, (char *)&save_special_chars);
  /* allow QUIT signal */
  (void) sigsetmask(0);
  (void) kill(getpid(), 3);
  exit_game();
}
#endif

/*ARGSUSED*/
#ifdef USG
void signal_save_no_core(int sig)
{
  error_sig = sig;
  error_code = 0;
  save_char(FALSE, TRUE);
  exit_game();
}
#else
void signal_save_no_core(int sig, int code, struct sigcontext *scp)
{
  error_sig = sig;
  error_code = code;
  save_char(FALSE, TRUE);
  exit_game();
}
#endif

/*ARGSUSED*/
#ifdef USG
void signal_ask_quit(int sig)
{
  char command;

  /* reset signal handler */
  (void) signal(sig, signal_ask_quit);
  if (get_com("Do you really want to quit?", &command))
    switch(command)
      {
      case 'y': case 'Y':
        if (character_generated)
	  upon_death();
        else
          exit_game();
	break;
      }
  find_flag = FALSE;
  if (search_flag)
    search_off();
  if (py.flags.rest > 0)
    rest_off();
  erase_line(msg_line, msg_line);
}
#else
void signal_ask_quit(int sig, int code, struct sigcontext *scp)
{
  char command;

  /* no need to reset signal handler */
  if (get_com("Do you really want to quit?", &command))
    switch(command)
      {
      case 'y': case 'Y':
	if (character_generated)
	  upon_death();
	else
	  exit_game();
	break;
      }
  find_flag = FALSE;
  if (search_flag)
    search_off();
  if (py.flags.rest > 0)
    rest_off();
  erase_line(msg_line, msg_line);
}
#endif

void no_controlz(void)
{
#ifdef USG
  /* no suspend signal */
#else
  suspend_handler = signal(SIGTSTP, SIG_IGN);
#endif
}

void controlz(void)
{
#ifdef USG
  /* no suspend signal */
#else
  (void) signal(SIGTSTP, suspend_handler);
#endif
}
