/*
 * Copyright (c) 1990-1992 Siemens Corporate Research, Inc.
 *                         and Rutgers University
 * All rights reserved.
 *
 $ $Log: libfile.c,v $
 * Revision 1.1.1.4  1995/01/08  21:03:11  xxzhang
 * new version of ptt from Siemens. Jan. 2, 1995
 *
 * Revision 1.25  1994/04/21  14:04:57  rita
 * Function pointer changes: Added statistics for fptr analysis.
 * Fixed fp = signal(X,foo) to make <*fp,foo> if <*fp,foo> existed
 * at call node.  Resolved fptr for signal when 2nd arg is a fptr
 * and not a function.
 *
 * Revision 1.24  1994/04/13  21:02:23  rita
 * Fixed abort,exit handling and miscellaneous fixes for signal.
 * Made -M command option to print must alias solution and heap statistics.
 *
 * Revision 1.23  1994/04/12  19:14:56  rita
 * 1. First attempt at signal implemented.  Calls to signal can be processed
 *    for call graph analysis if the second argument is a function name and
 *    not a function pointer name.
 * 2. Added another cond for killing RDs: point of redefinition
 *    must post dominate the location where RD was created.
 * 3. Implemented detection of assignments where must aliases containing lhs
 *    are modified by the assignment: Kill these must aliases and dont create
 *    new must aliases <<*lhs,*rhs>>.
 *
 * Revision 1.22  1994/02/21  18:00:14  rita
 * Added blackboxing for ttynam and some defs for signal.
 * Optimized on fptr processing to reduce the number of nodes that require
 * fptr processing: Nodes where rhs is not a fptr obj and lhs is not included
 * in fptr class.
 *
 * Revision 1.21  1994/02/17  16:34:44  landi
 * Bug fixes.
 *
 * Revision 1.20  1994/02/01  15:39:10  rita
 * Casting incorporated into call graph analysis with function pointers.
 * Fixed bug in reporting of "call_by_list" for functions.
 * Considered "rhs1" and "rhs2" in assignment statements.
 *
 * Revision 1.19  1993/12/09  20:37:56  rita
 * Improved the performance for must alias analysis when applied to non-heap
 * variables (Hemant's method).
 *
 * Revision 1.18  1993/12/01  21:10:30  rita
 * Removed some debugging code
 *
 * Revision 1.17  1993/12/01  19:35:14  rita
 * Fixed bug in topsorting algorithm
 *
 * Revision 1.16  1993/11/29  15:53:57  landi
 * Various changes.
 *
 * Revision 1.15  1993/10/22  17:32:01  rita
 * Fixed my_malloc related bugs in aliasing, must aliasing and reaching
 * definitions.  This improved quality for DU reporting.  Also, improved quality
 * for DU reporting for structures.
 *
 * Revision 1.14  1993/08/04  17:58:24  rita
 * Resolved function pointer references for call graph analyzer tool.
 *
 * Revision 1.13  1993/07/01  20:38:19  rita
 * Fixed some bugs with Bill and made some changes to accommodate the
 * processing of the +F option with the updated ptt.
 *
 * Revision 1.12  1993/05/27  18:24:57  rita
 * Fixed must alias loop verification error.
 * Fixed bug in blackboxing that came about after code consolidation.
 * Fixed bug in reaching definitions that incorrectly walked structure fields when
 * the +m option was used.
 *
 * Revision 1.11  1993/02/02  19:35:36  rita
 * 1. Modified Tactic interface method for reporting locations for
 *    operand definitions and uses.
 * 2. Fixed memory leaks.
 *
 * Revision 1.19  1993/01/27  16:19:52  rita
 * 1. Fixed memory leak in blackbox.c.
 * 2. Changed parameter profiles of some blackboxed functions in
 *    libfile.c
 *    This avoids type mismatch errors when doing must alias analysis.
 * 3. Created extension to blackbox operation ADD_ALIASES so that
 *    <*$$, _iob[]> will be recognized to create aliases between
 *    <*func_name(), _iob[node_num_of_func_call]>.
 *
 * Revision 1.18  1993/01/15  14:59:49  landi
 * Made scanf a read/mod of stdin
 *
 * Revision 1.17  1993/01/14  20:26:11  rita
 * Added parsing for _iob[0], _iob[1], _iob[2].
 * Revision 1.9  1993/01/15  18:31:51  landi
 * Forget to change scanf in these two.
 *
 * Revision 1.18  1993/01/15  14:59:49  landi
 * Made scanf a read/mod of stdin
 *
 * Revision 1.17  1993/01/14  20:26:11  rita
 * Added parsing for _iob[0], _iob[1], _iob[2].
 * Fixed bugs relating to type mismatches.
 * Fixed memory leak in must_def_table.
 *
 * Revision 1.7  1993/01/11  15:48:30  rita
 * Changed library files to properly profile realloc() as returning
 * a char*.
 *
 * Changed blackbox.c to remove a warning message in the case when a
 * lookup on a heap variable indicates that the heap variable already
 * exists.  This situation can legally happen in alias_at_exit_implies.
 *
 * Revision 1.6  1993/01/05  21:03:12  rita
 * Implemented intraprocedural must_alias so as to remove spurious reaching
 * definitions of *__XXheap.  ptt will do this if the +A option is set in
 * $ALIAS_FLAGS.  This will not removed spurious reaching definitions of
 * non heap variables.  To accomplish this, set the +R option in $ALIAS_FLAGS.
 *
 * Revision 1.5  1992/10/14  17:16:02  rita
 * Modified the blackboxing operations for wrefresh() and ungetc()
 *
 * Revision 1.4  1992/10/13  18:13:13  rita
 * 1.  Profiled and optimized blackboxing code.
 * 2.  Profiled remaining blackboxing functions with the
 *     appropriate blackboxing operations in
 *        libfile.c
 *        libfile_wo_heap.c
 *        libfile_wo_iob.c
 *        libfile_wo_heap_wo_iob.c
 * 3.  Implemented OBJ_USES operation
 *
 * Revision 1.3  1992/10/06  15:20:49  rita
 * Make changes for blackboxing optimizations
 * Updates libfiles.
 *
 * Revision 1.2  1992/09/24  16:54:11  rita
 * Completed implementation of blackboxing operation HEAP_DEFS.
 * All calls to malloc(), calloc(), etc. whether they are direct
 * or indirect calls (via call to my_malloc()) will have new
 * heap variables created appropriately.
 *
 * Direct calls to blackboxed functions malloc(), calloc(), etc.
 * will have the new heap variables generated at the function call
 * node.  This is consistent with blackboxed functions that define
 * an arguments.
 *
 * Indirect calls to malloc(), calloc() that are issued from a user's
 * my_malloc() function will have the new heap variables generated at
 * the function return node.  This is consistent with non-blackboxed
 * functions that return a global value.
 *
 * Revision 1.1  1992/09/16  15:23:18  rita
 * Added libfile.c which contains the dummy definitions for
 * library functions and the blackboxing operations for all
 * library functions currently blackboxed.
 *
 *
 *   Dummy definitions of library functions 
 *
 *   these are necessary so that data flow is able to process
 *   the files
 */

#define bool char

struct _win_st {
	short		_cury, _curx;
	short		_maxy, _maxx;
	short		_begy, _begx;
	short		_flags;
	short		_ch_off;
	bool		_clear;
	bool		_leave;
	bool		_scroll;
	char		_y[1][1];
	short		*_firstch;
	short		*_lastch;
	struct _win_st	*_nextp, *_orig;
};
struct _win_st stdscr[2];

char *_ttyname[];

#define stdin (&_iob[0])
#define stdout (&_iob[1])
#define stderr (&_iob[2])

struct	_iobuf {
	int	_cnt;
	unsigned char *_ptr;
	unsigned char *_base;
	int	_bufsiz;
	short	_flag;
	char	_file;	
    } _iob[];

#define FILE struct _iobuf

struct	tm {
	int	tm_sec;
	int	tm_min;
	int	tm_hour;
	int	tm_mday;
	int	tm_mon;
	int	tm_year;
	int	tm_wday;
	int	tm_yday;
	int	tm_isdst;
	char	*tm_zone;
	long	tm_gmtoff;
} _localtimeVal[1],_gmtimeVal[1];

struct	dirent {
	long		d_off;		/* offset of next disk dir entry */
	unsigned long	d_fileno;	/* file number of entry */
	unsigned short	d_reclen;	/* length of this record */
	unsigned short	d_namlen;	/* length of string in d_name */
	char		d_name[255+1];	/* name (up to MAXNAMLEN + 1) */
} _readdirVal[1];


struct _dirdesc {
        int     dd_fd;
        long    dd_loc;
        long    dd_size;
        long    dd_bsize;
        long    dd_off;
        char    *dd_buf;
}  _opendirVal[1];

struct hostent {
        char    *h_name;        
        char    **h_aliases;    
        int     h_addrtype;     
        int     h_length;       
        char    **h_addr_list;  
} _gethostbynameVal[1];

struct  servent {
        char    *s_name;        
        char    **s_aliases;    
        int     s_port;         
        char    *s_proto;       
} _getservbynameVal[1];

struct passwd {
        char    *pw_name;
        char    *pw_passwd;
        long    pw_uid;
        long    pw_gid;
        char    *pw_age;
        char    *pw_comment;
        char    *pw_gecos;
        char    *pw_dir;
        char    *pw_shell;
} _getpwentVal[1];

/* For signal */

void handler0(){};
void (*SA[1])();

char _getenvVal[10];
char _tgetstrVal[1];

/* If this name is changed then change function is_heap_temp */
char __heap[];

void crt0()
{

  static char _argv[][5] = { "arg1", "arg2"};
  static char _envp[][10] = { "argenvp1", "argenvp2"};
  main(1,_argv,_envp);
}

#if 1

int    _filbuf() {}
int    _flsbuf() {}
int    _assert() {}

int    abort() {}
int    abs() {}
int    accept() {}
int    access() {}
int    alarm() {}
char*  asctime() {}
int    atof() {}
int    atoi() {}
int    atol() {}
int    bcmp() {}
int    bcopy() {}
int    bind() {}
int    bzero() {}
char*  calloc() {}
int    cbreak() {}
int    ceil() {}
int    cfree() {}
int    chdir() {}
int    chmod() {}
int    chown() {}
int    close() {}
int    closedir() {}
int    connect() {}
int    cos() {}
int    creat() {}
char*  crypt() {}
char*  ctime() {}
int    def_prog_mode() {}
int    dup() {}
int    dup2() {}
int    echo() {}
int    endwin() {}
int    execl() {}
int    execv() {}
int    execvp() {}
int    exit() {}
int    _exit() {}
int    exp() {}
int    fabs() {}
int    fclose() {}

int    fcntl() {}
FILE*  fdopen() {}
int    feof() {}
int    ferror() {}
int    clearerr() {}
int    fileno() {}
int    fflush() {}
char   fgetc() {}
char*  fgets() {}
int    flock() {}
int    floor() {}
int    fmod() {}
FILE*  fopen() {}
int    fork() {}
int    fprintf() {}
int    fputc() {}
int    fputs() {}
int    fread() {}
int    free() {}
FILE*  freopen() {}
int    fscanf() {}
int    fseek() {}
int    fstat() {}
int    ftruncate() {}
int    fwrite() {}
int    ftell() {}
int    ftime() {}
int    getc() {}
int    getchar() {}
int    gets() {}
char*   getenv() {}
int    getdents() {}
char*  getpwent() {}
int    geteuid() {}
int    getgid() {}
char*  gethostbyaddr() {}
struct hostent*  gethostbyname() {}
char*  getpwuid() {}
int    gettimeofday() {}
char*  getlogin() {}
int    getpid() {}
int    getopt() {}
int    getpagesize() {}
int    getrusage() {}
int    getitimer() {}

struct servent*  getservbyname() {}
int    getuid() {}
char*  getwd() {}
struct tm*  gmtime() {}
int    gtty() {}
int    htons() {}
char*  index() {}
int    inet_addr() {}
int    initscr() {}
int    initscr32() {}
int	clearok() {}
char*  initstate() {}
int    ioctl() {}
int    isalpha() {}
int    isatty() {}
int    iscntrl() {}
int    isdigit() {}
int    islower() {}
int    isprint() {}
int    isspace() {}
int    isupper() {}
int    keypad() {}
int    kill() {}
double ldexp(x,n) double x; int n; {}
int    link() {}
int    listen() {}
struct tm*  localtime() {}
int    log() {}
int    log10() {}
int    longjmp() {}
int    lrand48() {}
int    lseek() {}
int    lstat() {}
char*  malloc() {}
int    memcmp() {}
char*  memcpy() {}
char*  memset() {}
int    mkdir() {}
char*  mktemp() {}
int    nl() {}
int    nocbreak() {}
int    noecho() {}
int    nonl() {}
int    open() {}
struct _dirdesc*  opendir() {}
int    pause() {}
int    pclose() {}
int    perror() {}
int    pipe() {}
FILE*  popen() {}
int    pow() {}

int    putc() {}
int    putchar() {}
int    puts() {}
int    printf() {}
int    printw() {}
int    qsort() {}
int    rand() {}
int    random() {}
int    re_comp() {}
int    re_exec() {}
int    read() {}
struct dirent*  readdir() {}
int    raw() {}
int    readlink() {}
char*  realloc() {}
int    recv() {}
char*  regcmp() {}
char*  regex() {}
int    remove() {}
int    rename() {}
int    reset_prog_mode() {}
int    reset_shell_mode() {}
int    rewind() {}
int    rindex() {}
int    rmdir() {}
int    scandir() {}
int    scanf() {}
int    seekdir() {}
int    select() {}
int    sendto() {}
int    setbuf() {}
int    setbuffer() {}
int    setgid() {}
int    setitimer() {}
int    setjmp() {}
int    setlinebuf() {}
int    setstate() {}
int    setuid() {}
void   (*signal())() {}  
int    sigblock() {}

int    sigpause() {}
int    sigsetmask() {}
int    sleep() {}
int    socket() {}

int    sprintf() {}
int    sqrt() {}
int    srand() {}
int    srand48() {}
int    srandom() {}
int    sscanf() {}
int    stat() {}
char*  strchr() {}
int    strcmp() {}
int    strcasecmp() {}
int    strncasecmp() {}
char*  strcat() {}
char*  strpbrk() {}
char*  strcpy() {}
char*  strdup() {}
char*  strerror() {}
int    strftime() {}
int    strlen() {}
int    strncmp() {}
int    strncpy() {}
char*  strrchr() {}
int    strtol() {}
int    stty() {}
int    syscall() {}
int    system() {}
char *getcwd(buf, size) char *buf; int size; {}
int    tgetent() {}
int    tgetflag() {}
int    tgetnum() {}
char*  tgetstr() {}
int    tgoto() {}
int    time() {}
int    times() {}
int    tolower() {}
int    toupper() {}
int    tputs() {}
char*  ttyname() {}
int    tzset() {}
int    umask() {}
int    ungetc() {}
int    unlink() {}
int    uname() {}
int    utime() {}
int    utimes() {}
int    vfork() {}
int    vfprintf() {}
int    addch() {}
int    waddch() {}
int    w32addch() {}
int    addstr() {}
int    mvaddstr() {}
int    waddstr() {}
int    wait() {}
int    waitpid() {}
int    clear() {}
int    wclear() {}
int    clrtobot() {}
int    wclrtobot() {}
int    clrtoeol() {}
int    wclrtoeol() {}
int    getch() {}
int    wgetch() {}
int    winch() {}
int    move() {}
int    wmove() {}
int    refresh() {}
int    wrefresh() {}
int    write() {}
int    standend() {}
int    wstandend() {}
int    standout() {}
int    wstandout() {}

#endif

/*  BLACKBOXED LIBRARY FUNCTIONS */
/*=

atoi() getpid() select() strcmp() strlen() system()
bind() strncmp() write() unlink() atol() atof() 
re_comp() re_exec() ftell() access() fork()
execl() getlogin() execvp() chdir() 
creat() chmod() chown() utime()
execv() ioctl() ldexp() asctime() inet_addr()
iscntrl() isdigit() islower() isprint() isspace() isupper() 
strcasecmp() sendto() connect()
gethostbyaddr() setbuf() setbuffer() link() vfork() seekdir()
fcntl() feof() ferror() fileno() gtty() stty() memcmp() mkdir() bcmp()
ftruncate() rmdir() getopt() rand() umask() tzset()
lrand48() 
utimes() remove() rename() strncasecmp() 
setlinebuf() syscall() getuid()
flock() wrefresh() setuid() geteuid() setstate()
sigsetmask() wstandout() wstandend() tgoto() sigblock()
sigpause() setgid() getgid() tgetnum() 
_filbuf() _flsbuf() _assert() pause() 
keypad() winch() nonl() noecho() raw() reset_prog_mode() 
def_prog_mode() reset_shell_mode() cbreak() nocbreak() nl() echo() 
getpagesize() clear() mvaddstr() refresh() getch() move() 
clrtobot() clrtoeol() addstr() addch() standend() standout()
alarm():;

abs() ceil() close() isatty() log10() random() toupper() 
sqrt() srandom() htons() ungetc() fabs() fmod() dup() sleep()
isalpha() 
exp() tolower() srand() srand48() log() cos() floor():
ARG_USES($1);

clearerr():
ARG_DEFS(*$1);

getenv():
ADD_ALIASES(*$$, _getenvVal[])
ARG_USES(*$1);

getpwent():
ADD_ALIASES(*$$, _getpwentVal[]);

localtime():
ADD_ALIASES(*$$, _localtimeVal[]);

listen() pow() dup2():
ARG_USES($1, $2);

lseek() socket():
ARG_USES($1, $2, $3);

tputs() setjmp() longjmp() kill():
NONE(..);

setitimer():
ARG_DEFS(*$2, *$3);

signal():
SIG_HANDLERS($1, $2);

accept():
ARG_DEFS(*$2, *$3);

bcopy():
ARG_DEFS(*$1);

bzero():
ARG_DEFS(*$1)
ARG_USES($2);

calloc() malloc() strdup():
HEAP_DEFS(*__heap);

realloc():
HEAP_DEFS(*__heap)
ADD_ALIASES(*$$,*$1);

cfree() free():
ARG_KILL_DEFS(*$1);

exit() _exit() abort():
ARG_KILL_DEFS(..);

fclose() pclose():
ARG_KILL_DEFS(*$1);

fflush() sprintf() rewind() fseek() fread():
ARG_DEFS(*$1);

fgets():
ADD_ALIASES(*$$, *$1)
ARG_DEFS(*$1)
ARG_USE_DEF(*$3);

fdopen() fopen() freopen() popen():
OBJ_USE_DEF(_iob[])
ADD_ALIASES(*$$, _iob[]);

fprintf() vfprintf():
ARG_USE_DEF(*$1);

fputc():
ARG_USE_DEF(*$2);

fputs():
ARG_USES($1)
ARG_USE_DEF(*$2);

fscanf():
ARG_DEFS(*$3..)
ARG_USE_DEF(*$1);

ftime():
ARG_DEFS(*$1);

fwrite():
ARG_DEFS(*$4);

getc() fgetc():
ARG_USE_DEF(*$1);

getchar():
OBJ_USE_DEF(_iob[0]);

getrusage():
ARG_USES($1)
ARG_DEFS(*$2);

getdents():
ARG_DEFS(*$2);

gethostbyname():
ADD_ALIASES(*$$,_gethostbynameVal[]);

getservbyname():
ADD_ALIASES(*$$,_getservbynameVal[]);

gettimeofday():
ARG_DEFS(*$1, *$2);

gets():
OBJ_USE_DEF(_iob[0])
ARG_DEFS(*$1)
ADD_ALIASES(*$$, *$1);

getitimer():
ARG_DEFS(*$2);

getwd():
ARG_DEFS(*$1)
ADD_ALIASES(*$$, *$1);

gmtime():
ADD_ALIASES(*$$,_gmtimeVal[]);

index() rindex():
ARG_USES($2)
ADD_ALIASES(*$$, *$1);

initscr() initscr32() endwin() printw():
OBJ_USE_DEF(stdscr[]);

clearok():
ARG_DEFS(*$1);

initstate():
ARG_DEFS(*$2);

memcpy() memset():
ARG_DEFS(*$1)
ADD_ALIASES(*$$, *$1);

mktemp():
ARG_DEFS(*$1)
ADD_ALIASES(*$$, *$1);

perror():
OBJ_USE_DEF(_iob[2]);

pipe():
ARG_DEFS(*$1);

printf() puts():
OBJ_USE_DEF(_iob[1]);

putc():
ARG_USES($1)
OBJ_USE_DEF(_iob[1]);

putchar():
OBJ_USE_DEF(_iob[1]);

qsort():
ARG_DEFS(*$1);

read():
ARG_DEFS(*$2);

readlink():
ARG_DEFS(*$2);

recv():
ARG_DEFS(*$2);

scandir():
ARG_DEFS(*$2);

scanf():
ARG_DEFS(*$2..)
OBJ_USE_DEF(_iob[0]);

sscanf():
ARG_DEFS(*$3..);

stat() fstat() lstat():
ARG_DEFS(*$2);

strcat() strcpy() strncpy():
ARG_DEFS(*$1)
ADD_ALIASES(*$$, *$1);

strerror():
HEAP_DEFS(*__heap);

strftime():
ARG_DEFS(*$1);

strpbrk():
ARG_USES(*$1..)
ADD_ALIASES(*$$, *$1);

strrchr() strchr():
ARG_USES($2)
ADD_ALIASES(*$$, *$1);

tgetent():
ARG_DEFS(*$1)
ARG_USES(*$2);

tgetflag():
ARG_USES(*$1);

tgetstr():
ADD_ALIASES(*$$, _tgetstrVal[])
ARG_DEFS(*$2);

time() times():
ARG_DEFS(*$1);

ttyname():
OBJ_DEFS(_ttyname[])
ADD_ALIASES(*$$, _ttyname[]);

wait() wmove():
ARG_DEFS(*$1);

waitpid():
ARG_DEFS(*$2);

wclear() waddstr() wgetch() wclrtoeol() wclrtobot() waddch() w32addch():
ARG_USE_DEF(*$1);

crypt():
HEAP_DEFS(*__heap);

ctime():
HEAP_DEFS(*__heap);

getcwd():
ADD_ALIASES(*$$, *$1)
HEAP_DEFS(*__heap);

opendir():
ADD_ALIASES(*$$,_opendirVal[]);

readdir():
ADD_ALIASES(*$$, _readdirVal[]);

closedir():
ARG_KILL_DEFS(*$1);

regcmp():
HEAP_DEFS(*__heap);

regex():
ADD_ALIASES(*$$,*$2);

strtol():
ADD_ALIASES(*$2,$1);

uname():
ARG_DEFS(*$1);

=*/


int yywrap() {}
